import logging

from cattle_grid.account.models import ActorForAccount
from cattle_grid.activity_pub.models import Actor
from cattle_grid.model.exchange import (
    UpdateAction,
    UpdateActionType,
    UpdateIdentifierAction,
    RenameActorAction,
    UpdatePropertyValueAction,
)

from .property_value import handle_update_property_value, handle_delete_property_value
from .identifiers import (
    handle_create_identifier,
    handle_update_identifier,
    handle_add_identifier,
)

logger = logging.getLogger(__name__)


async def handle_rename_actor(actor: Actor, action: RenameActorAction) -> None:
    actor_for_account = await ActorForAccount.get_or_none(actor=actor.actor_id)
    if not actor_for_account:
        logger.warning("ActorForAccount does not exist for actor %s", actor.actor_id)
        return

    actor_for_account.name = action.name
    await actor_for_account.save()


async def handle_actor_action(actor: Actor, action: UpdateAction) -> bool:
    """Handles individual actions from the UpdateActor
    method"""
    match action.action:
        case UpdateActionType.add_identifier:
            action = UpdateIdentifierAction.model_validate(action.model_dump())
            await handle_add_identifier(actor, action)
            return True
        case UpdateActionType.create_identifier:
            action = UpdateIdentifierAction.model_validate(action.model_dump())
            await handle_create_identifier(actor, action)
            return True
        case UpdateActionType.update_identifier:
            action = UpdateIdentifierAction.model_validate(action.model_dump())
            await handle_update_identifier(actor, action)
            return True
        case UpdateActionType.rename:
            rename_action = RenameActorAction.model_validate(action.model_dump())
            await handle_rename_actor(actor, rename_action)
            return False
        case UpdateActionType.update_property_value:
            cast_action = UpdatePropertyValueAction.model_validate(action.model_dump())
            await handle_update_property_value(actor, cast_action)
            return True
        case UpdateActionType.remove_property_value:
            cast_action = UpdatePropertyValueAction.model_validate(action.model_dump())
            await handle_delete_property_value(actor, cast_action)
            return True

    return False



from cattle_grid.model.account import (
    NameAndVersion,
    InformationResponse,
    ActorInformation,
)
from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.version import __version__


from cattle_grid.account.models import Account, ActorForAccount, ActorStatus
from cattle_grid.account.permissions import allowed_base_urls


def protocol_and_backend():
    protocol = NameAndVersion(name="CattleDrive", version="0.1.0")
    backend = NameAndVersion(name="cattle_grid", version=__version__)

    return dict(protocol=protocol, backend=backend)


def actor_to_information(actor: ActorForAccount) -> ActorInformation:
    """Transform ActorForAccount to its information ActorInformation

    ```pycon
    >>> actor = ActorForAccount(actor="http://base.example/actor", name="Alice")
    >>> actor_to_information(actor)
    ActorInformation(id='http://base.example/actor', name='Alice')

    ```
    """
    return ActorInformation(id=actor.actor, name=actor.name)


async def create_information_response(
    account: Account, method_information: list[MethodInformationModel]
) -> InformationResponse:
    await account.fetch_related("actors")

    actor_ids = [
        actor_to_information(x)
        for x in account.actors
        if x.status == ActorStatus.active
    ]
    base_urls = await allowed_base_urls(account)

    return InformationResponse(
        account_name=account.name,
        base_urls=base_urls,
        actors=actor_ids,
        **protocol_and_backend(),
        method_information=method_information,
    )



from unittest.mock import AsyncMock
from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.activity_pub.models import StoredActivity
from cattle_grid.model.processing import StoreActivityMessage

from .store_activity import store_activity_subscriber


async def test_store_activity(test_actor):  # noqa
    broker = AsyncMock()
    activity = {
        "actor": test_actor.actor_id,
        "type": "Activity",
        "to": ["http://remote.test/actor"],
    }
    msg = StoreActivityMessage(actor=test_actor.actor_id, data=activity)

    await store_activity_subscriber(msg, test_actor, broker)

    assert 1 == await StoredActivity.filter().count()

    broker.publish.assert_awaited_once()



from faststream import Context
from faststream.rabbit import RabbitBroker

from bovine.activitystreams.utils import recipients_for_object

from cattle_grid.model import SharedInboxMessage
from cattle_grid.activity_pub.models import Actor, Following
from cattle_grid.activity_pub.enqueuer import enqueue_from_inbox
from cattle_grid.dependencies.globals import global_container


async def handle_shared_inbox_message(
    message: SharedInboxMessage,
    broker: RabbitBroker = Context(),
):
    """
    This method is used to handle incoming messages from the shared inbox.
    """

    recipients = recipients_for_object(message.data)
    sender = message.data.get("actor")

    if sender is None:
        return

    local_actor_ids = {
        x.actor_id for x in await Actor.filter(actor_id__in=recipients).all()
    }
    following_actor_ids = {
        x.actor.actor_id
        for x in await Following.filter(following=sender, accepted=True)
        .prefetch_related("actor")
        .all()
    }

    for actor in local_actor_ids | following_actor_ids:
        await enqueue_from_inbox(
            broker, global_container.internal_exchange, actor, message.data
        )



"""ActivityPub related functionality"""

import logging
from fastapi import APIRouter, HTTPException, Header
from fastapi.responses import JSONResponse

from pydantic import BaseModel
from typing import Annotated
from bovine.activitystreams import OrderedCollection

from cattle_grid.activity_pub.actor import (
    actor_to_object,
    followers_for_actor,
    following_for_actor,
)
from cattle_grid.activity_pub.models import Actor, ActorStatus

logger = logging.getLogger(__name__)

ap_router = APIRouter()


class APHeaders(BaseModel):
    """Headers every request should have. These should be added by the remote proxy."""

    x_cattle_grid_requester: str
    """URI of the actor making the request"""
    x_ap_location: str
    """URI of the resource being retrieved"""


ActivityPubHeaders = Annotated[APHeaders, Header()]


class ActivityResponse(JSONResponse):
    """Response that ensures the content-type is
    "application/activity+json"
    """

    media_type = "application/activity+json"


def validate_actor(actor: Actor | None) -> Actor:
    if actor is None:
        raise HTTPException(404)
    if actor.status == ActorStatus.deleted:
        raise HTTPException(410)

    return actor


async def ensure_not_blocked(actor: Actor, retriever: str) -> None:
    await actor.refresh_from_db()
    await actor.fetch_related("blocking")

    logger.info("List of blocked actors")
    logger.info([x.blocking for x in actor.blocking])

    if any(block.blocking == retriever and block.active for block in actor.blocking):
        raise HTTPException(403)

    return None


@ap_router.get("/actor/{id_str}", response_class=ActivityResponse)
async def actor_profile(id_str, headers: ActivityPubHeaders):
    """Returns the actor"""
    logger.debug("Request for actor at %s", headers.x_ap_location)
    actor = validate_actor(await Actor.get_or_none(actor_id=headers.x_ap_location))
    await ensure_not_blocked(actor, headers.x_cattle_grid_requester)

    await actor.fetch_related("identifiers")

    result = actor_to_object(actor)
    return result


@ap_router.get("/outbox/{id_str}", response_class=ActivityResponse)
async def outbox(id_str, headers: ActivityPubHeaders):
    """Returns an empty ordered collection as outbox"""
    actor = validate_actor(await Actor.get_or_none(outbox_uri=headers.x_ap_location))
    await ensure_not_blocked(actor, headers.x_cattle_grid_requester)

    return OrderedCollection(id=headers.x_ap_location).build()


@ap_router.get("/following/{id_str}", response_class=ActivityResponse)
async def following(id_str, headers: ActivityPubHeaders):
    """Returns the following"""
    actor = validate_actor(await Actor.get_or_none(following_uri=headers.x_ap_location))
    await ensure_not_blocked(actor, headers.x_cattle_grid_requester)

    following = await following_for_actor(actor)
    return OrderedCollection(id=headers.x_ap_location, items=following).build()


@ap_router.get("/followers/{id_str}", response_class=ActivityResponse)
async def followers(id_str, headers: ActivityPubHeaders):
    """Returns the followers"""
    actor = validate_actor(await Actor.get_or_none(followers_uri=headers.x_ap_location))
    await ensure_not_blocked(actor, headers.x_cattle_grid_requester)

    followers = await followers_for_actor(actor)
    return OrderedCollection(id=headers.x_ap_location, items=followers).build()



import logging
from typing import Any, Dict
from contextlib import asynccontextmanager

from aerich import Command
from tortoise import Tortoise
from .config import load_settings

logger = logging.getLogger(__name__)

TORTOISE_ORM = {
    "connections": {"default": "postgres://postgres:pass@cattle_grid_db"},
    "apps": {
        "aerich": {
            "models": ["aerich.models"],
        },
        "ap_models": {
            "models": ["cattle_grid.activity_pub.models", "cattle_grid.account.models"]
        },
        "auth_models": {
            "models": ["cattle_grid.auth.model"],
        },
    },
}

# app_list = ["aerich", "ap_models", "gateway_models", "auth_models"]

app_list = ["ap_models"]


def tortoise_config(db_uri: str) -> Dict[str, Any]:
    TORTOISE_ORM["connections"] = {"default": db_uri}
    return TORTOISE_ORM


@asynccontextmanager
async def database(db_uri: str = "sqlite://:memory:", generate_schemas: bool = False):
    """Opens the connection to the database using tortoise"""
    await Tortoise.init(config=tortoise_config(db_uri))
    if generate_schemas:
        await Tortoise.generate_schemas()

    try:
        yield
    finally:
        await Tortoise.close_connections()


def determine_migration_location():
    return "/".join(__file__.split("/")[:-1] + ["migrations"])


#
# Not sure the following function is what one wants
# ... it means pinning one database migrations work
# ... for
#


async def upgrade(config=load_settings()) -> None:
    for app in app_list:
        command = Command(
            tortoise_config=tortoise_config(config.db_uri),  # type: ignore
            location=determine_migration_location(),
            app=app,
        )

        await command.init()
        migrated = await command.upgrade(run_in_transaction=False)

        if not migrated:
            logger.info("Performed no migrations for %s", app)
        else:
            for version_file in migrated:
                logger.info("Preformed migration %s for %s", version_file, app)


async def migrate(config=load_settings(), name: str | None = None) -> None:
    for app in app_list:
        try:
            location = determine_migration_location()
            print(location)
            command = Command(
                tortoise_config=tortoise_config(config.db_uri),  # type:ignore
                location=location,
                app=app,
            )

            if name is None:
                name = "update"

            await command.init()
            await command.migrate(name=name)
        except Exception as e:
            print(e)


async def run_with_database(config, coro):
    try:
        async with database(db_uri=config.db_uri):
            await coro
    except Exception as e:
        logger.exception(e)



import asyncio
import pytest
import click

from unittest.mock import MagicMock

from click.testing import CliRunner

from cattle_grid.database import database
from cattle_grid.testing.fixtures import *  # noqa

from .account import account_with_username_password, list_permissions
from .cli import add_account_commands, new_account, modify_permissions


@pytest.fixture
def db_uri(tmp_path):
    return "sqlite://" + str(tmp_path / "test.db")


@pytest.fixture
def cli(db_uri):
    config = MagicMock()

    config.db_uri = db_uri

    @click.group()
    @click.pass_context
    def main(ctx):
        ctx.ensure_object(dict)
        ctx.obj["config"] = config

    add_account_commands(main)

    return main


@pytest.fixture
def create_database(db_uri):
    async def run_method():
        async with database(db_uri=db_uri, generate_schemas=True):
            await asyncio.sleep(0.1)
        from tortoise import Tortoise

        await Tortoise.close_connections()

    asyncio.run(run_method())


def test_new_account_cli(create_database, cli, tmp_path):
    runner = CliRunner()
    result = runner.invoke(cli, ["account", "new", "user", "pass"])

    assert result.exit_code == 0


async def test_new_account():
    await new_account("user", "pass", permission=["one"])

    account = await account_with_username_password("user", "pass")

    assert account
    await account.fetch_related("permissions")
    permissions = list_permissions(account)

    assert permissions == ["one"]


async def test_modify_permissions():
    await new_account("user", "pass", permission=["one"])
    await modify_permissions("user", ["two"], ["one"])

    account = await account_with_username_password("user", "pass")

    assert account
    await account.fetch_related("permissions")
    permissions = list_permissions(account)

    assert permissions == ["two"]



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "publicidentifier" ADD "status" VARCHAR(10)NOT NULL DEFAULT 'verified';"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "publicidentifier" DROP COLUMN "status";"""



import pytest
from unittest.mock import AsyncMock, MagicMock

from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.future import select

from cattle_grid.testing.fixtures import *  # noqa

from . import lifespan, simple_storage_publish_activity, simple_storage_publish_object
from .config import SimpleStorageConfiguration
from .message_types import PublishActivity, PublishObject
from .models import StoredActivity, StoredObject

activity_stub = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "AnimalSound",
    "actor": "http://alice.example",
    "to": ["http://bob.example"],
    "content": "moo",
}

object_stub = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "Note",
    "attributedTo": "http://alice.example",
    "to": ["http://bob.example"],
    "content": "moo",
}


@pytest.fixture(autouse=True)
async def sql_engine(sql_engine_for_tests):
    async with lifespan(sql_engine_for_tests):
        yield sql_engine_for_tests


@pytest.mark.parametrize(
    "actor,activity",
    [
        ("http://bob.example", activity_stub),
        (
            "http://alice.example",
            {**activity_stub, "id": "http://alice.example/activity1"},
        ),
    ],
)
async def test_simple_storage_publish_activity_errors(actor, activity):
    msg = PublishActivity(
        actor=actor,
        data=activity,
    )

    with pytest.raises(ValueError):
        await simple_storage_publish_activity(
            msg, AsyncMock(), MagicMock(), AsyncMock()
        )


async def test_simple_storage_activity(sql_engine, test_actor):
    msg = PublishActivity(actor="http://alice.example", data=activity_stub)

    broker = AsyncMock()
    config = SimpleStorageConfiguration(prefix="/simple/storage/")

    async_session = async_sessionmaker(sql_engine, expire_on_commit=False)

    async with async_session() as session:
        await simple_storage_publish_activity(
            msg, config, broker=broker, session=session
        )
        await session.commit()

    broker.publish.assert_awaited_once()

    async with async_session() as session:
        result = (await session.scalars(select(StoredActivity))).all()

        assert len(result) == 1

        assert result[0].data.get("id")


@pytest.mark.parametrize(
    "actor,object",
    [
        ("http://bob.example", object_stub),
        (
            "http://alice.example",
            {**object_stub, "id": "http://alice.example/activity1"},
        ),
    ],
)
async def test_simple_storage_publish_object_errors(actor, object):
    msg = PublishObject(
        actor=actor,
        data=object,
    )

    with pytest.raises(ValueError):
        await simple_storage_publish_object(msg, AsyncMock(), MagicMock(), AsyncMock())


async def test_simple_storage_object(sql_engine, test_actor):
    obj = object_stub.copy()
    obj["attributedTo"] = test_actor.actor_id
    msg = PublishObject(actor=test_actor.actor_id, data=obj)

    broker = AsyncMock()
    config = SimpleStorageConfiguration(prefix="/simple/storage/")

    async_session = async_sessionmaker(sql_engine, expire_on_commit=False)

    async with async_session() as session:
        await simple_storage_publish_object(
            msg, config, actor=test_actor, broker=broker, session=session
        )
        await session.commit()

    broker.publish.assert_awaited_once()

    async with async_session() as session:
        result = (await session.scalars(select(StoredActivity))).all()

        assert len(result) == 1

        assert result[0].data.get("id")

        result = (await session.scalars(select(StoredObject))).all()

        assert len(result) == 1

        assert result[0].data.get("id")



import pytest
from unittest.mock import AsyncMock

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model import SharedInboxMessage
from cattle_grid.activity_pub.models import Following

from .shared_inbox import handle_shared_inbox_message


async def test_no_recipients_unknown_actor_nothing_happens():
    broker = AsyncMock()
    message = SharedInboxMessage(
        data={"actor": "https://remote.test/actor", "type": "Activity"}
    )

    await handle_shared_inbox_message(message, broker)

    broker.publish.assert_not_awaited()


@pytest.mark.parametrize(
    "to_func",
    [
        lambda x: x.actor_id,
        lambda x: [x.actor_id],
        lambda x: [x.actor_id, "http://other.test/actor"],
    ],
)
async def test_addressed_to_actor(test_actor, to_func):
    activity = {
        "actor": "http://remote.test/actor",
        "type": "Activity",
        "to": to_func(test_actor),
    }

    broker = AsyncMock()
    message = SharedInboxMessage(data=activity)

    await handle_shared_inbox_message(message, broker)

    broker.publish.assert_awaited_once()


async def test_actor_is_following(test_actor):
    remote_actor = "http://remote.test/actor"
    activity = {
        "actor": remote_actor,
        "type": "Activity",
        "to": "http://remote.test/actor",
    }

    await Following.create(
        actor=test_actor, following=remote_actor, accepted=1, request="x"
    )

    broker = AsyncMock()
    message = SharedInboxMessage(data=activity)

    await handle_shared_inbox_message(message, broker)

    broker.publish.assert_awaited_once()



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.account.account import create_account, add_permission
from cattle_grid.model.account import InformationResponse

from cattle_grid.account.models import ActorForAccount, ActorStatus

from .info import create_information_response


@pytest.fixture
async def test_admin_account():
    account = await create_account("test_account", "test_password")
    assert account
    await add_permission(account, "admin")

    return account


async def test_create_information_response(test_admin_account):
    response = await create_information_response(test_admin_account, [])

    assert isinstance(response, InformationResponse)


async def test_create_information_response_actors(test_admin_account):
    await ActorForAccount.create(
        account=test_admin_account,
        actor="http://host.test/actor/active",
        status=ActorStatus.active,
    )
    await ActorForAccount.create(
        account=test_admin_account,
        actor="http://host.test/actor/deleted",
        status=ActorStatus.deleted,
    )
    await test_admin_account.fetch_related("actors")

    response = await create_information_response(test_admin_account, [])

    actors = response.actors
    assert len(actors) == 1

    assert actors[0].id.endswith("active")



from typing import Annotated

from fastapi import HTTPException, Depends
from fastapi.security import OAuth2PasswordBearer

from cattle_grid.account.models import AuthenticationToken, Account

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/signin")


async def get_current_account(token: Annotated[str, Depends(oauth2_scheme)]) -> Account:
    from_db = await AuthenticationToken.get_or_none(token=token).prefetch_related(
        "account"
    )

    if from_db is None:
        raise HTTPException(401)

    return from_db.account


CurrentAccount = Annotated[Account, Depends(get_current_account)]
"""Annotation for the current account"""



import os
import pytest

from .auth.model import RemoteIdentity
from .database import database, upgrade, determine_migration_location


async def test_database():
    async with database(generate_schemas=True):
        await RemoteIdentity.filter().count()


@pytest.mark.skip("Requires Postgres sql")
async def test_upgrade():
    await upgrade(db_uri="sqlite://cattle_grid.db")


def test_determine_migration_location():
    location = determine_migration_location()

    contents = os.listdir(location)

    assert "ap_models" in contents



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "actorforaccount" ADD "name" VARCHAR(255)NOT NULL DEFAULT 'NO NAME';
        COMMENT ON COLUMN "actorforaccount"."actor" IS 'The uri of the actor';"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "actorforaccount" DROP COLUMN "name";
        COMMENT ON COLUMN "actorforaccount"."actor" IS NULL;"""



def skip_method_information(routing_key):
    """Method information is not created for subscribtions to incoming.* or outgoing.*

    as only cattle_grid should publish to these
    routing keys

    ```pycon
    >>> skip_method_information("incoming.test")
    True

    >>> skip_method_information("outgoing.test")
    True

    >>> skip_method_information("test")
    False

    ```
    """

    if routing_key.startswith("incoming.") or routing_key.startswith("outgoing."):
        return True

    return False



"""

Data model used to describe ActivityPub related objects

"""

from typing import Dict, Any

from tortoise import fields
from tortoise.models import Model

from enum import StrEnum, auto


class ActorStatus(StrEnum):
    """Represents the status of the actor"""

    active = auto()
    deleted = auto()


class PublicIdentifierStatus(StrEnum):
    """Represents the status of the public identifier"""

    unverified = auto()
    """This identifier could not be verified"""

    verified = auto()
    """This identifier was verified"""

    owned = auto()
    """This is an identifier owned by the cattle_grid instance"""


class InboxLocation(Model):
    """Describes the location of an inbox. Used to send
    ActivityPub Activities addressed to the actor to the
    corresponding inbox.

    This information is also collected for remote actors.
    """

    id = fields.IntField(primary_key=True)
    actor = fields.CharField(max_length=255, unique=True)
    """The id of the remote actor"""
    inbox = fields.CharField(max_length=255)
    """The inbox of the remote actor"""


class Actor(Model):
    """Actors administrated by cattle_grid"""

    id = fields.IntField(primary_key=True)
    actor_id = fields.CharField(max_length=255, unique=True)
    """The id of the actor"""

    inbox_uri = fields.CharField(max_length=255, unique=True)
    """The uri of the inbox"""
    outbox_uri = fields.CharField(max_length=255, unique=True)
    """The uri of the outbox"""
    following_uri = fields.CharField(max_length=255, unique=True)
    """The uri of the following collection"""
    followers_uri = fields.CharField(max_length=255, unique=True)
    """The uri of the followers collection"""

    preferred_username = fields.CharField(max_length=255, null=True)
    """The preferred username, used as the username part of the
    acct-uri of the actor, i.e. `acct:${preferred_username}@domain`.
    See [RFC 7565 The 'acct' URI Scheme](https://www.rfc-editor.org/rfc/rfc7565.html)."""

    public_key_name = fields.CharField(max_length=255)
    """The name given to the public key, i.e. the id will be
    `${actor_id}#${public_key_name}."""
    public_key = fields.TextField()
    """The public key"""

    automatically_accept_followers = fields.BooleanField()
    """Set to true to indicate cattle_grid should automatically
    accept follow requests"""
    profile: Dict[str, Any] = fields.JSONField()  # type:ignore
    """Additional profile values"""

    status = fields.CharEnumField(ActorStatus, default=ActorStatus.active)
    """Represents the status of the actor"""

    followers: fields.ReverseRelation["Follower"]
    following: fields.ReverseRelation["Following"]
    identifiers: fields.ReverseRelation["PublicIdentifier"]


class Follower(Model):
    """The people that follow the actor"""

    id = fields.IntField(primary_key=True)

    actor: fields.ForeignKeyRelation[Actor] = fields.ForeignKeyField(
        "ap_models.Actor", related_name="followers"
    )

    follower = fields.CharField(max_length=255)
    request = fields.CharField(max_length=255)
    accepted = fields.BooleanField()


class Following(Model):
    """The people the actor is following"""

    id = fields.IntField(primary_key=True)

    actor: fields.ForeignKeyRelation[Actor] = fields.ForeignKeyField(
        "ap_models.Actor", related_name="following"
    )

    following = fields.CharField(max_length=255)
    request = fields.CharField(max_length=255)
    accepted = fields.BooleanField()


class PublicIdentifier(Model):
    """Public identifiers"""

    id = fields.IntField(primary_key=True)

    actor: fields.ForeignKeyRelation[Actor] = fields.ForeignKeyField(
        "ap_models.Actor", related_name="identifiers"
    )
    """The actor the public key belongs to"""

    name = fields.CharField(max_length=255)
    """name of public identifier"""
    identifier = fields.CharField(max_length=255, unique=True)
    """The public identifier, e.g. an acct-uri"""
    preference = fields.IntField(default=0)
    """Determines the order of identifiers in the `identifiers` field of the actor profile"""

    status = fields.CharEnumField(
        PublicIdentifierStatus, default=PublicIdentifierStatus.verified
    )
    """Represents the status of the public identifier"""


class Credential(Model):
    """The secrets of the actor"""

    id = fields.IntField(primary_key=True)

    actor_id = fields.CharField(max_length=255)
    identifier = fields.CharField(max_length=255)

    secret = fields.TextField()


class StoredActivity(Model):
    """cattle_grid generates activities under some
    circumstances (see FIXME). These will be stored
    in this table"""

    id = fields.CharField(max_length=255, primary_key=True)

    actor: fields.ForeignKeyRelation[Actor] = fields.ForeignKeyField("ap_models.Actor")
    """The actor this activity orginates from"""

    data: dict = fields.JSONField()  # type: ignore
    """The activity"""

    published = fields.DatetimeField()
    """When the activity was published"""


class Blocking(Model):
    """The people the actor is blocking"""

    id = fields.IntField(primary_key=True)

    actor: fields.ForeignKeyRelation[Actor] = fields.ForeignKeyField(
        "ap_models.Actor", related_name="blocking"
    )

    blocking = fields.CharField(max_length=255)
    request = fields.CharField(max_length=255)
    active = fields.BooleanField()



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        CREATE TABLE IF NOT EXISTS "aerich" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "version" VARCHAR(255) NOT NULL,
    "app" VARCHAR(100) NOT NULL,
    "content" JSONB NOT NULL
);
CREATE TABLE IF NOT EXISTS "actor" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "actor_id" VARCHAR(255) NOT NULL UNIQUE,
    "inbox_uri" VARCHAR(255) NOT NULL UNIQUE,
    "outbox_uri" VARCHAR(255) NOT NULL UNIQUE,
    "following_uri" VARCHAR(255) NOT NULL UNIQUE,
    "followers_uri" VARCHAR(255) NOT NULL UNIQUE,
    "preferred_username" VARCHAR(255),
    "public_key_name" VARCHAR(255) NOT NULL,
    "public_key" TEXT NOT NULL,
    "automatically_accept_followers" BOOL NOT NULL,
    "profile" JSONB NOT NULL,
    "status" VARCHAR(7) NOT NULL  DEFAULT 'active'
);
COMMENT ON COLUMN "actor"."status" IS 'active: active\ndeleted: deleted';
COMMENT ON TABLE "actor" IS 'Actors administrated by cattle_grid';
CREATE TABLE IF NOT EXISTS "blocking" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "blocking" VARCHAR(255) NOT NULL,
    "request" VARCHAR(255) NOT NULL,
    "active" BOOL NOT NULL,
    "actor_id" INT NOT NULL REFERENCES "actor" ("id") ON DELETE CASCADE
);
COMMENT ON TABLE "blocking" IS 'The people the actor is blocking';
CREATE TABLE IF NOT EXISTS "credential" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "actor_id" VARCHAR(255) NOT NULL,
    "identifier" VARCHAR(255) NOT NULL,
    "secret" TEXT NOT NULL
);
COMMENT ON TABLE "credential" IS 'The secrets of the actor';
CREATE TABLE IF NOT EXISTS "follower" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "follower" VARCHAR(255) NOT NULL,
    "request" VARCHAR(255) NOT NULL,
    "accepted" BOOL NOT NULL,
    "actor_id" INT NOT NULL REFERENCES "actor" ("id") ON DELETE CASCADE
);
COMMENT ON TABLE "follower" IS 'The people that follow the actor';
CREATE TABLE IF NOT EXISTS "following" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "following" VARCHAR(255) NOT NULL,
    "request" VARCHAR(255) NOT NULL,
    "accepted" BOOL NOT NULL,
    "actor_id" INT NOT NULL REFERENCES "actor" ("id") ON DELETE CASCADE
);
COMMENT ON TABLE "following" IS 'The people the actor is following';
CREATE TABLE IF NOT EXISTS "inboxlocation" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "actor" VARCHAR(255) NOT NULL UNIQUE,
    "inbox" VARCHAR(255) NOT NULL
);
COMMENT ON TABLE "inboxlocation" IS 'Describes the location of an inbox. Used to send';
CREATE TABLE IF NOT EXISTS "publicidentifier" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "name" VARCHAR(255) NOT NULL,
    "identifier" VARCHAR(255) NOT NULL UNIQUE,
    "preference" INT NOT NULL  DEFAULT 0,
    "actor_id" INT NOT NULL REFERENCES "actor" ("id") ON DELETE CASCADE
);
COMMENT ON TABLE "publicidentifier" IS 'Public identifiers';
CREATE TABLE IF NOT EXISTS "storedactivity" (
    "id" VARCHAR(255) NOT NULL  PRIMARY KEY,
    "data" JSONB NOT NULL,
    "published" TIMESTAMPTZ NOT NULL,
    "actor_id" INT NOT NULL REFERENCES "actor" ("id") ON DELETE CASCADE
);
COMMENT ON TABLE "storedactivity" IS 'cattle_grid generates activities under some';
CREATE TABLE IF NOT EXISTS "account" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "name" VARCHAR(255) NOT NULL,
    "password_hash" VARCHAR(255) NOT NULL
);
COMMENT ON TABLE "account" IS 'Represents an Account';
CREATE TABLE IF NOT EXISTS "actorforaccount" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "actor" VARCHAR(255) NOT NULL,
    "account_id" INT NOT NULL REFERENCES "account" ("id") ON DELETE CASCADE
);
CREATE TABLE IF NOT EXISTS "authenticationtoken" (
    "token" VARCHAR(64) NOT NULL  PRIMARY KEY,
    "account_id" INT NOT NULL REFERENCES "account" ("id") ON DELETE CASCADE
);
CREATE TABLE IF NOT EXISTS "remoteidentity" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "key_id" VARCHAR(512) NOT NULL UNIQUE,
    "controller" VARCHAR(512) NOT NULL,
    "public_key" VARCHAR(1024) NOT NULL
);
COMMENT ON TABLE "remoteidentity" IS 'Represents the information about a remote identity';"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        """



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "actorgroup" ALTER COLUMN "id" ADD GENERATED ALWAYS AS IDENTITY; """


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """ALTER TABLE "actorgroup" ALTER COLUMN "id" DROP IDENTITY;
        """



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        CREATE TABLE IF NOT EXISTS "account" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "name" VARCHAR(255) NOT NULL,
    "password_hash" VARCHAR(255) NOT NULL
);
COMMENT ON TABLE "account" IS 'Represents an Account';
        CREATE TABLE IF NOT EXISTS "actorforaccount" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "actor" VARCHAR(255) NOT NULL,
    "account_id" INT NOT NULL REFERENCES "account" ("id") ON DELETE CASCADE
);
        CREATE TABLE IF NOT EXISTS "authenticationtoken" (
    "token" VARCHAR(64) NOT NULL PRIMARY KEY,
    "account_id" INT NOT NULL REFERENCES "account" ("id") ON DELETE CASCADE
);
        CREATE TABLE IF NOT EXISTS "permission" (
    "id" SERIAL NOT NULL PRIMARY KEY,
    "name" VARCHAR(255) NOT NULL,
    "account_id" INT NOT NULL REFERENCES "account" ("id") ON DELETE CASCADE
);"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        DROP TABLE IF EXISTS "actorforaccount";
        DROP TABLE IF EXISTS "authenticationtoken";
        DROP TABLE IF EXISTS "account";
        DROP TABLE IF EXISTS "permission";"""



import asyncio
import logging
import aio_pika
from contextlib import asynccontextmanager
from uuid import uuid4

from cattle_grid.dependencies.fastapi import Broker
from cattle_grid.model.account import EventType

logger = logging.getLogger(__name__)


@asynccontextmanager
async def queue_for_connection(connection, username: str, event_type: EventType):
    async with connection:
        routing_key = f"receive.{username}.{event_type.value}"

        channel = await connection.channel()
        await channel.set_qos(prefetch_count=1)

        exchange = await channel.declare_exchange(
            "amq.topic", aio_pika.ExchangeType.TOPIC, durable=True
        )

        queue = await channel.declare_queue(
            f"queue-{username}-{uuid4()}",
            durable=False,
            auto_delete=True,
            exclusive=True,
        )
        await queue.bind(exchange, routing_key=routing_key)

        yield queue


async def enqueue_events(
    asyncio_queue, connection, username: str, event_type: EventType
):
    async with queue_for_connection(connection, username, event_type) as queue:
        async with queue.iterator() as iterator:
            try:
                async for message in iterator:
                    async with message.process():
                        message_body = message.body.decode()
                        await asyncio_queue.put(message_body)
            except Exception as e:
                logger.exception(e)


def get_message_streamer(broker: Broker, timeout: float = 5):
    if broker._connection is None:
        raise RuntimeError("Broker not connected")

    connection: aio_pika.RobustConnection = broker._connection

    async def stream_messages(username: str, event_type: EventType):
        queue = asyncio.Queue()
        task = asyncio.create_task(
            enqueue_events(queue, connection, username, event_type)
        )
        try:
            while True:
                try:
                    async with asyncio.timeout(timeout):
                        result = await queue.get()
                except TimeoutError:
                    logger.debug("Sending heartbeat")
                    result = None
                yield result
        except asyncio.CancelledError:
            logger.info("cancelling message queue for %s %s", username, event_type)
            task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass
        except Exception as e:
            logger.exception(e)

    return stream_messages



import logging

from bovine.activitystreams.utils.property_value import from_key_value


from cattle_grid.activity_pub.models import Actor

from cattle_grid.model.exchange import UpdatePropertyValueAction

logger = logging.getLogger(__name__)


def find_key_in_attachments(attachments: list, key: str) -> int | None:
    for idx, attachment in enumerate(attachments):
        if not isinstance(attachment, dict):
            continue
        if attachment.get("type") != "PropertyValue":
            continue
        if attachment.get("name") == key:
            return idx

    return None


async def handle_update_property_value(
    actor: Actor, action: UpdatePropertyValueAction
) -> None:
    current_attachments = actor.profile.get("attachment", [])

    current_index = find_key_in_attachments(current_attachments, action.key)

    if current_index is None:
        current_attachments.append(from_key_value(action.key, action.value))
    else:
        current_attachments[current_index] = from_key_value(action.key, action.value)

    actor.profile["attachment"] = current_attachments

    await actor.save()


async def handle_delete_property_value(
    actor: Actor, action: UpdatePropertyValueAction
) -> None:
    current_attachments = actor.profile.get("attachment", [])

    current_index = find_key_in_attachments(current_attachments, action.key)

    if current_index is not None:
        del current_attachments[current_index]

    actor.profile["attachment"] = current_attachments

    logger.info(current_attachments)

    await actor.save()



import asyncio
import pytest

from click.testing import CliRunner

from cattle_grid.database import database

from .__main__ import main


@pytest.fixture
def db_uri(tmp_path):
    return "sqlite://" + str(tmp_path / "test.db")


@pytest.fixture
def create_database(db_uri):
    async def run_method():
        async with database(db_uri=db_uri, generate_schemas=True):
            await asyncio.sleep(0.5)
        from tortoise import connections

        await connections.close_all()

        await asyncio.sleep(0.3)

    asyncio.run(run_method())


def test_list(db_uri, create_database):
    runner = CliRunner(env={"CATTLE_GRID_DB_URI": db_uri})
    result = runner.invoke(main, ["list"])

    print(result.output)

    assert result.exit_code == 0



import pytest

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model import ActivityMessage

from cattle_grid.activity_pub.models import Follower, Following

from .util import update_recipients_for_collections


@pytest.fixture
def message(test_actor):  # noqa
    return ActivityMessage(actor=test_actor.actor_id, data={})


async def test_update_recipients_for_collection(message, test_actor):  # noqa
    recipients = {"http://remote.test"}

    result = await update_recipients_for_collections(message, recipients)

    assert result == recipients


async def test_update_recipients_for_collection_with_followers(
    message,
    test_actor,  # noqa
):
    recipients = {"http://remote.test", test_actor.followers_uri}

    await Follower.create(
        actor=test_actor, follower="http://follower.test", accepted=True, request="none"
    )

    result = await update_recipients_for_collections(message, recipients)

    assert result == {"http://remote.test", "http://follower.test"}


async def test_update_recipients_for_collection_for_following_followers(
    message,
    test_actor,  # noqa
):
    recipients = {
        "http://remote.test",
        test_actor.followers_uri,
        test_actor.following_uri,
    }

    await Follower.create(
        actor=test_actor, follower="http://follower.test", accepted=True, request="none"
    )
    await Following.create(
        actor=test_actor,
        following="http://following.test",
        accepted=True,
        request="none",
    )

    result = await update_recipients_for_collections(message, recipients)

    assert result == {"http://remote.test", "http://follower.test"}


async def test_update_recipients_for_collection_for_self_delete(
    message,
    test_actor,  # noqa
):
    recipients = {
        "http://remote.test",
        test_actor.followers_uri,
        test_actor.following_uri,
    }

    await Follower.create(
        actor=test_actor, follower="http://follower.test", accepted=True, request="none"
    )
    await Following.create(
        actor=test_actor,
        following="http://following.test",
        accepted=True,
        request="none",
    )

    message.data = {
        "type": "Delete",
        "actor": test_actor.actor_id,
        "object": test_actor.actor_id,
    }

    result = await update_recipients_for_collections(message, recipients)

    assert result == {
        "http://remote.test",
        "http://follower.test",
        "http://following.test",
    }



import asyncio

import click

from cattle_grid.activity_pub.models import Actor, ActorStatus
from cattle_grid.account.models import (
    ActorForAccount,
    ActorStatus as ActorStatusForAccount,
)
from cattle_grid.database import run_with_database


async def list_actors(deleted: bool = False):
    result = await Actor.filter().all()

    if deleted:
        result = [x for x in result if x.status == ActorStatus.deleted]

    for x in result:
        print(x.actor_id)

    if len(result) == 0:
        print("No actors")


async def prune_actors():
    await Actor.filter(status=ActorStatus.deleted).delete()
    await ActorForAccount.filter(status=ActorStatusForAccount.deleted).delete()


def add_actors_to_cli_as_group(main):
    @main.group()
    def actor():
        """Used to manage actors"""

    add_to_cli(actor)


def add_to_cli(main):
    @main.command("list")  # type: ignore
    @click.option(
        "--deleted", is_flag=True, default=False, help="Only list deleted actors"
    )
    @click.pass_context
    def list_actors_command(ctx, deleted):
        asyncio.run(run_with_database(ctx.obj["config"], list_actors(deleted)))

    @main.command("prune")  # type: ignore
    @click.pass_context
    def prune_actors_command(ctx):
        asyncio.run(run_with_database(ctx.obj["config"], prune_actors()))



import logging

from enum import Enum
from typing import Annotated, List

from fastapi import HTTPException, Request, Response, APIRouter, Header
from starlette.datastructures import MutableHeaders

from pydantic import BaseModel

from bovine.activitystreams import Actor
from bovine.utils import webfinger_response, parse_fediverse_handle
from bovine.crypto.signature_checker import SignatureChecker
from bovine.types.jrd import JrdData

from urllib.parse import urlparse
from contextlib import asynccontextmanager

from cattle_grid.activity_pub.server.router import ActivityResponse
from cattle_grid.config.auth import AuthConfig
from cattle_grid.dependencies.globals import global_container

from .public_key_cache import PublicKeyCache
from .util import config_to_bovine_actor, check_block
from .http_util import should_serve, ContentType

logger = logging.getLogger(__name__)


class ReverseProxyHeaders(BaseModel):
    """Headers set by the reverse proxy"""

    x_original_method: str = "get"
    """The original used method"""

    x_original_uri: str | None = None
    """The original request uri"""

    x_original_host: str | None = None
    """The original used host"""

    x_forwarded_proto: str = "http"
    """The protocol being used"""


def create_auth_router(
    config: AuthConfig, tags: List[str | Enum] = ["auth"]
) -> APIRouter:
    """Adds the authorization endpoint to the app"""

    bovine_actor = config_to_bovine_actor(config)
    public_key_cache = PublicKeyCache(bovine_actor)
    signature_checker = SignatureChecker(
        public_key_cache.from_cache, skip_digest_check=True
    )

    actor_path = str(urlparse(config.actor_id).path)
    webfinger_jrd = webfinger_response(config.actor_acct_id, config.actor_id)

    username, _ = parse_fediverse_handle(config.actor_acct_id.removeprefix("acct:"))
    actor_object = Actor(
        id=config.actor_id,
        type="Service",
        public_key=config.public_key,
        preferred_username=username,
        public_key_name="mykey",
    ).build()

    @asynccontextmanager
    async def lifespan(app):
        if global_container.session:
            await bovine_actor.init(session=global_container.session)
            yield
        else:
            async with global_container.session_lifecycle():
                await bovine_actor.init(session=global_container.session)
                yield

    router = APIRouter(lifespan=lifespan, tags=tags)

    @router.get(
        actor_path,
        response_class=ActivityResponse,
    )
    async def handle_get_actor():
        """Returns the actor profile of the
        fetch actor used to retrieve public keys, e.g.

        ```json
        {
            "type": "Service",
            "id": "https://your-domain.example/cattle_grid_actor",
            ...
        }
        ```
        """
        return actor_object

    @router.get("/.well-known/webfinger")
    async def webfinger(resource: str) -> JrdData:
        """If resource is the actor corresponding to the actor fetching
        public keys, returns the corresponding Jrd. Otherwise returns
        not found"""
        if resource != config.actor_acct_id:
            raise HTTPException(404)
        return webfinger_jrd

    @router.get(
        "/auth",
        responses={
            401: {"description": "The signature was invalid"},
            403: {"description": "Request was blocked"},
        },
    )
    async def verify_signature(
        request: Request,
        response: Response,
        reverse_proxy_headers: Annotated[ReverseProxyHeaders, Header()],
    ):
        """Takes the request and checks signature. If signature check
        fails a 401 is returned. If the domain the public key belongs
        to is blocked, a 403 is returned.

        If the request is valid. The controller corresponding to
        the signature is set in the response header `X-CATTLE-GRID-REQUESTER`.

        Note: More headers than the ones listed below can be used
        to verify a signature.
        """
        headers = MutableHeaders(request.headers)

        logger.debug(headers)

        servable_content_types = should_serve(headers.get("accept"))

        if "signature" not in headers:
            if ContentType.html in servable_content_types:
                response.headers["x-cattle-grid-should-serve"] = "html"
                return ""
            elif ContentType.other in servable_content_types:
                response.headers["x-cattle-grid-should-serve"] = "other"
                return ""
            elif config.require_signature_for_activity_pub:
                raise HTTPException(401)
            else:
                return ""

        if reverse_proxy_headers.x_original_host:
            headers["Host"] = reverse_proxy_headers.x_original_host

        url = f"{reverse_proxy_headers.x_forwarded_proto}://{reverse_proxy_headers.x_original_host}{reverse_proxy_headers.x_original_uri}"

        logger.debug("Treating request as to url %s", url)

        controller = await signature_checker.validate_signature(
            reverse_proxy_headers.x_original_method.lower(),
            url,
            dict(headers.items()),
            None,
        )

        logger.debug("Got controller %s", controller)

        if controller:
            if check_block(config.domain_blocks, controller):
                logger.info("Blocked a request by %s", controller)
                raise HTTPException(403)
            response.headers["x-cattle-grid-requester"] = controller

            logger.debug("Got requester %s", controller)

            return ""

        logger.info(
            "invalid signature for request to %s => access denied",
            request.headers.get("X-Original-Uri", ""),
        )

        raise HTTPException(401)

    return router



import pytest
from cattle_grid.account.account import create_account, add_actor_to_account
from cattle_grid.activity_pub.actor import create_actor
from cattle_grid.database import database as with_database

from cattle_grid.config.auth import new_auth_config, save_auth_config
from cattle_grid.dependencies.globals import global_container


@pytest.fixture(autouse=True)
async def database_for_tests():
    """Fixture so that the database is initialized"""
    async with with_database(db_uri="sqlite://:memory:", generate_schemas=True):
        yield


@pytest.fixture
async def test_account():
    """Fixture to create an account"""
    return await create_account("alice", "alice", permissions=["admin"])


@pytest.fixture
async def test_actor():
    """Fixture to create an actor"""
    return await create_actor("http://localhost/ap")


@pytest.fixture
async def test_actor_with_account(test_account):
    """Fixture to create an actor with an account"""
    actor = await create_actor("http://localhost/ap")
    await add_actor_to_account(test_account, actor, name="test_fixture")

    return actor


@pytest.fixture
def auth_config_file(tmp_path):
    config = new_auth_config(actor_id="http://localhost/actor_id", username="actor")

    filename = tmp_path / "auth_config.toml"

    config.domain_blocks = set(["blocked.example"])

    save_auth_config(filename, config)

    return filename


@pytest.fixture(autouse=True, scope="session")
def loaded_config():
    global_container.load_config()


@pytest.fixture()
async def sql_engine_for_tests():
    """Provides the sql engine (as in memory sqlite) for tests"""
    async with global_container.alchemy_database(
        "sqlite+aiosqlite:///:memory:", echo=False
    ) as engine:
        yield engine



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        CREATE TABLE IF NOT EXISTS  "actorgroup"  (
    "id" integer NOT NULL,
    "name" character varying(255) NOT NULL,
    "actor_id" integer NOT NULL
);
ALTER TABLE ONLY public.actorgroup
    ADD CONSTRAINT actorgroup_pkey PRIMARY KEY (id);


ALTER TABLE ONLY public.actorgroup
    ADD CONSTRAINT actorgroup_actor_id_fkey FOREIGN KEY (actor_id) REFERENCES public.actorforaccount(id) ON DELETE CASCADE;


"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        DROP TABLE "actorgroup";"""



from enum import StrEnum, auto

from tortoise import fields
from tortoise.models import Model


class ActorStatus(StrEnum):
    """Status actors can have for an account"""

    active = auto()
    deleted = auto()


class Account(Model):
    """Represents an Account"""

    id = fields.IntField(primary_key=True)

    name = fields.CharField(max_length=255)
    """The account name"""

    password_hash = fields.CharField(max_length=255)
    """The hashed password"""

    actors: fields.ReverseRelation["ActorForAccount"]
    """Actors associated with this account"""

    token: fields.ReverseRelation["AuthenticationToken"]
    """Authentication tokens for this account"""

    permissions: fields.ReverseRelation["Permission"]
    """Permissions the account has"""

    meta_information = fields.JSONField(
        default={},
    )
    """Additional information about the account"""


class ActorForAccount(Model):
    """Lists the actors associated with the account"""

    id = fields.IntField(primary_key=True)

    account: fields.ForeignKeyRelation[Account] = fields.ForeignKeyField(
        "ap_models.Account", related_name="actors"
    )
    actor = fields.CharField(max_length=255, description="The uri of the actor")
    """The URI of the actor"""
    name = fields.CharField(
        max_length=255,
        description="human readable name for the actor",
        default="NO NAME",
    )
    """internal name of the actor, used for the human behind the account to identify it"""
    status = fields.CharEnumField(ActorStatus, default=ActorStatus.active)
    """status"""

    groups: fields.ReverseRelation["ActorGroup"]
    """Groups the actor is a member of"""


class AuthenticationToken(Model):
    token = fields.CharField(max_length=64, primary_key=True)

    account: fields.ForeignKeyRelation[Account] = fields.ForeignKeyField(
        "ap_models.Account", related_name="tokens"
    )


class Permission(Model):
    id = fields.IntField(primary_key=True)

    name = fields.CharField(max_length=255)
    account: fields.ForeignKeyRelation[Account] = fields.ForeignKeyField(
        "ap_models.Account", related_name="permissions"
    )


class ActorGroup(Model):
    """Groups the actor is a member of"""

    id = fields.IntField(primary_key=True)
    name = fields.CharField(max_length=255)
    """Name of the group"""
    actor: fields.ForeignKeyRelation[ActorForAccount] = fields.ForeignKeyField(
        "ap_models.ActorForAccount", related_name="groups"
    )



import logging

from behave import when, then
from behave.api.async_step import async_run_until_complete

from bovine.activitystreams import factories_for_actor_object

from cattle_grid.activity_pub.models import Actor
from cattle_grid.activity_pub.actor import is_valid_requester

from cattle_grid.database import database

logger = logging.getLogger(__name__)


@when('"{alice}" creates an object addressed to "{recipient}"')
def object_addressed_to(context, alice, recipient):
    alice_actor = context.actors[alice]
    _, object_factory = factories_for_actor_object(alice_actor)

    if recipient == "public":
        context.object = object_factory.note(content="moo").as_public().build()
    elif recipient == "followers":
        context.object = object_factory.note(content="moo").as_followers().build()
    else:
        context.object = object_factory.note(
            content="moo", to={context.actors[recipient].get("id")}
        ).build()


@then('"{bob}" is "{state}" to view this object')
@async_run_until_complete
async def check_allowed(context, bob, state):
    bob_id = context.actors[bob].get("id")

    async with database(db_uri="postgres://postgres:pass@cattle_grid_db"):
        alice = await Actor.get_or_none(actor_id=context.object.get("attributedTo"))

        is_valid = await is_valid_requester(bob_id, alice, context.object)

    if is_valid:
        assert state == "authorized"
    else:
        assert state == "unauthorized"



from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.account.models import ActorForAccount
from cattle_grid.model.exchange import UpdateAction, UpdateActionType

from .actor_update import handle_actor_action


async def test_handle_actor_action_rename(test_actor, test_account):
    actor_for_account = await ActorForAccount.create(
        actor=test_actor.actor_id, account=test_account
    )

    action = UpdateAction(
        action=UpdateActionType.rename,
        name="new name",  # type:ignore
    )

    await handle_actor_action(test_actor, action)

    await actor_for_account.refresh_from_db()

    assert actor_for_account.name == "new name"



from unittest.mock import AsyncMock, MagicMock
from contextlib import asynccontextmanager
from asyncstdlib.itertools import islice

from .streaming import get_message_streamer


async def test_get_message_streamer():
    broker = MagicMock()

    @asynccontextmanager
    async def connection():
        yield AsyncMock()

    broker._connection = connection()

    streamer = get_message_streamer(AsyncMock(), 0.1)

    count = 0
    count_target = 3

    async for x in islice(streamer("username", "event_type"), count_target):
        count = count + 1

    assert count == count_target



import logging
import secrets

from typing import List
from urllib.parse import urljoin, urlparse

from bovine import BovineActor
from bovine.activitystreams import Actor as AsActor, factories_for_actor_object
from bovine.activitystreams.utils import recipients_for_object, is_public
from bovine.activitystreams.utils.property_value import property_value_context
from bovine.crypto import generate_rsa_public_private_key
from bovine.types import Visibility


from cattle_grid.activity_pub.activity import actor_deletes_themselves
from cattle_grid.activity_pub.models import (
    Actor,
    PublicIdentifier,
    Credential,
    Follower,
    Following,
    ActorStatus,
    Blocking,
)

from .identifiers import (
    determine_preferred_username,
    collect_identifiers_for_actor,
    identifier_in_list_exists,
)
from .helper import endpoints_object_from_actor_id

logger = logging.getLogger(__name__)


class DuplicateIdentifierException(Exception):
    """Raised if an identifier already exists and one tries to create an actor with it"""


class ActorNotFound(Exception):
    """Raised if an actor is not found"""


def new_url(base_url: str, url_type: str) -> str:
    token = secrets.token_urlsafe(16)
    return urljoin(base_url, f"{url_type}/{token}")


def compute_acct_uri(base_url: str, preferred_username: str):
    """Computes the acct uri

    ```pycon
    >>> compute_acct_uri("http://host.example/somewhere", "alice")
    'acct:alice@host.example'

    ```

    """
    host = urlparse(base_url).hostname

    return f"acct:{preferred_username}@{host}"


async def create_actor(
    base_url: str,
    preferred_username: str | None = None,
    identifiers: dict = {},
    profile: dict = {},
):
    """Creates a new actor in the database"""

    public_key, private_key = generate_rsa_public_private_key()
    public_key_name = "legacy-key-1"
    actor_id = new_url(base_url, "actor")

    if preferred_username:
        if "webfinger" in identifiers:
            raise ValueError("webfinger key set in identifiers")
        identifiers = {
            **identifiers,
            "webfinger": compute_acct_uri(base_url, preferred_username),
        }

    if "activitypub_id" not in identifiers:
        identifiers = {**identifiers, "activitypub_id": actor_id}

    identifier_already_exists = await identifier_in_list_exists(
        list(identifiers.values())
    )

    if identifier_already_exists:
        raise DuplicateIdentifierException("identifier already exists")

    actor = await Actor.create(
        actor_id=actor_id,
        inbox_uri=new_url(base_url, "inbox"),
        outbox_uri=new_url(base_url, "outbox"),
        following_uri=new_url(base_url, "following"),
        followers_uri=new_url(base_url, "followers"),
        public_key_name=public_key_name,
        public_key=public_key,
        profile=profile,
        automatically_accept_followers=False,
    )
    await Credential.create(
        actor_id=actor_id,
        identifier=f"{actor_id}#{public_key_name}",
        secret=private_key,
    )

    for name, identifier in identifiers.items():
        await PublicIdentifier.create(actor=actor, name=name, identifier=identifier)

    logging.info("Created actor with id '%s'", actor_id)

    await actor.fetch_related("identifiers")

    return actor


def actor_to_object(actor: Actor) -> dict:
    """Transform the actor to an object

    :params actor:
    :returns:
    """

    sorted_identifiers = collect_identifiers_for_actor(actor)

    preferred_username = determine_preferred_username(
        sorted_identifiers, actor.actor_id
    )
    attachments = actor.profile.get("attachment")
    result = AsActor(
        id=actor.actor_id,
        outbox=actor.outbox_uri,
        inbox=actor.inbox_uri,
        followers=actor.followers_uri,
        following=actor.following_uri,
        public_key=actor.public_key,
        public_key_name=actor.public_key_name,
        preferred_username=preferred_username,
        type=actor.profile.get("type", "Person"),
        name=actor.profile.get("name"),
        summary=actor.profile.get("summary"),
        url=actor.profile.get("url"),
        icon=actor.profile.get("image", actor.profile.get("icon")),
        properties={
            "attachment": attachments,
        },
    ).build(visibility=Visibility.OWNER)

    result["identifiers"] = sorted_identifiers
    result["endpoints"] = endpoints_object_from_actor_id(actor.actor_id)

    if attachments:
        result["@context"].append(property_value_context)

    return result


async def bovine_actor_for_actor_id(actor_id: str) -> BovineActor | None:
    """Uses the information stored in [Credential][cattle_grid.activity_pub.models.Credential] to construct a bovine actor

    :params actor_id:
    :returns:
    """
    credential = await Credential.get_or_none(actor_id=actor_id)

    if credential is None:
        return None

    return BovineActor(
        public_key_url=credential.identifier,
        actor_id=actor_id,
        secret=credential.secret,
    )


async def followers_for_actor(actor: Actor) -> List[str]:
    """Returns the list of accepted followers

    :param actor:
    :returns:
    """

    await actor.fetch_related("followers")
    return [x.follower for x in actor.followers if x.accepted]


async def following_for_actor(actor: Actor) -> List[str]:
    """Returns the list of accepted people to follow said actor.
    This is the following table.

    :param actor:
    :returns:
    """

    await actor.fetch_related("following")
    return [x.following for x in actor.following if x.accepted]


def update_for_actor_profile(actor: Actor) -> dict:
    """Creates an update for the Actor"""

    actor_profile = actor_to_object(actor)
    activity_factory, _ = factories_for_actor_object(actor_profile)

    return (
        activity_factory.update(actor_profile, followers=actor_profile["followers"])
        .as_public()
        .build()
    )


def delete_for_actor_profile(actor: Actor) -> dict:
    """Creates a delete activity for the Actor"""

    actor_profile = actor_to_object(actor)
    activity_factory, _ = factories_for_actor_object(actor_profile)

    result = (
        activity_factory.delete(
            actor_profile.get("id"), followers=actor_profile["followers"]
        )
        .as_public()
        .build()
    )

    result["cc"].append(actor_profile["following"])

    return result


async def delete_actor(actor: Actor):
    """Deletes an actor

    :param actor: Actor to be deleted
    """

    # await Credential.filter(actor_id=actor.actor_id).delete()
    await PublicIdentifier.filter(actor=actor).delete()

    actor.status = ActorStatus.deleted
    await actor.save()


async def remove_from_followers_following(actor_id_to_remove: str):
    """Removes actor_id from all occurring followers and following"""

    await Follower.filter(follower=actor_id_to_remove).delete()
    await Following.filter(following=actor_id_to_remove).delete()


async def update_recipients_for_actor(actor, recipients, self_delete=False):
    """Updates set of recipients by removing the followers and following collections, and replacing
    them with the actual sets.

    The following collecting is only allowed for self delete activities.
    """
    if actor.followers_uri in recipients:
        recipients = recipients - {actor.followers_uri} | set(
            await followers_for_actor(actor)
        )

        logger.info("Got recipients %s after handling followers", ", ".join(recipients))

    if actor.following_uri in recipients:
        recipients = recipients - {actor.following_uri}

        if self_delete:
            recipients = recipients | set(await following_for_actor(actor))
        else:
            logger.warning(
                "Actor '%s' included following collection in recipients where not allowed",
                actor.actor_id,
            )

    return recipients


async def is_valid_requester(requester: str, actor: Actor, obj: dict):
    """Checks if the requested is allowed to view the object"""

    blocked = await Blocking.get_or_none(blocking=requester, actor=actor, active=True)

    if blocked:
        return False

    if is_public(obj):
        return True

    recipients = recipients_for_object(obj)
    self_delete = actor_deletes_themselves(obj)

    recipients = await update_recipients_for_actor(actor, recipients, self_delete)

    valid_requesters = recipients

    if "actor" in obj:
        valid_requesters = valid_requesters | {obj["actor"]}
    if "attributedTo" in obj:
        valid_requesters = valid_requesters | {obj["attributedTo"]}

    return requester in valid_requesters


async def is_valid_requester_for_obj(requester: str, obj: dict):
    """Checks if the requested is allowed to view the object"""

    actor_id = obj.get("attributedTo")
    if actor_id is None:
        actor_id = obj.get("actor")
    if actor_id is None:
        raise ActorNotFound("Object does not have an actor or attributedTo")

    actor = await Actor.get_or_none(actor_id=actor_id)
    if actor is None:
        raise ActorNotFound("Actor not found")

    blocked = await Blocking.get_or_none(blocking=requester, actor=actor, active=True)

    if blocked:
        return False

    if is_public(obj):
        return True

    recipients = recipients_for_object(obj)
    self_delete = actor_deletes_themselves(obj)

    recipients = await update_recipients_for_actor(actor, recipients, self_delete)

    valid_requesters = recipients

    if "actor" in obj:
        valid_requesters = valid_requesters | {obj["actor"]}
    if "attributedTo" in obj:
        valid_requesters = valid_requesters | {obj["attributedTo"]}

    return requester in valid_requesters



from cattle_grid.testing.fixtures import sql_engine_for_tests  # noqa

from . import lifespan


async def test_lifespan(sql_engine_for_tests):  # noqa
    async with lifespan(sql_engine_for_tests):
        pass



import argon2
import logging
import re

from cattle_grid.config.settings import get_settings
from cattle_grid.activity_pub.models import Actor

from .models import Account, Permission, ActorStatus, ActorForAccount, ActorGroup

logger = logging.getLogger(__name__)

password_hasher = argon2.PasswordHasher()


class AccountAlreadyExists(Exception):
    pass


class InvalidAccountName(Exception):
    pass


class WrongPassword(Exception):
    pass


async def create_account(
    name: str,
    password: str,
    settings=get_settings(),
    permissions: list[str] = [],
    meta_information: dict[str, str] = {},
) -> Account | None:
    """Creates a new account for username and password"""
    if await Account.get_or_none(name=name):
        raise AccountAlreadyExists("Account already exists")

    if not re.match(settings.account.allowed_name_regex, name):  # type: ignore
        raise InvalidAccountName("Account name does not match allowed format")

    if name in settings.account.forbidden_names:  # type: ignore
        raise InvalidAccountName("Account name is forbidden")

    account = await Account.create(
        name=name,
        password_hash=password_hasher.hash(password),
        meta_information=meta_information,
    )
    for permission in permissions:
        await add_permission(account, permission)
    return account


async def account_with_username_password(name: str, password: str) -> Account | None:
    """Retrieves account for given username and password"""
    account = await Account.get_or_none(name=name)
    if account is None:
        return None

    try:
        password_hasher.verify(account.password_hash, password)
    except argon2.exceptions.VerifyMismatchError:
        logger.warning("Got wrong password for %s", name)
        return None

    # Implement rehash?
    # https://argon2-cffi.readthedocs.io/en/stable/howto.html

    return account


async def delete_account(
    name: str, password: str | None = None, force: bool = False
) -> None:
    """Deletes account for given username and password

    If password is wrong or account does not exist,
    raises a WrongPassword exception"""
    if force:
        account = await Account.get_or_none(name=name)
    else:
        if not password:
            raise ValueError("Password must be provided if not forcing deletion")
        account = await account_with_username_password(name, password)

    if account is None:
        raise WrongPassword(
            "Either the account does not exist or the password is wrong"
        )

    await account.fetch_related("actors")
    active_actors = [a for a in account.actors if a.status == ActorStatus.active]
    if len(active_actors) > 0:
        logger.warning(
            "Deleting account with active actors: %s", [a.actor for a in active_actors]
        )

    await account.delete()


async def add_permission(account: Account, permission: str) -> None:
    """Adds permission to account"""
    await Permission.create(account=account, name=permission)


async def remove_permission(account: Account, permission: str) -> None:
    """Removes permission from account"""
    p = await Permission.get_or_none(account=account, name=permission)

    if p:
        await p.delete()


def list_permissions(account: Account) -> list[str]:
    """Returns list of permissions for account"""
    return [p.name for p in account.permissions]


async def account_for_actor(actor: str) -> Account | None:
    """Given an actor_id returns the corresponding account or None"""
    actor_for_account = await ActorForAccount.get_or_none(actor=actor).prefetch_related(
        "account"
    )

    if not actor_for_account:
        return None

    return actor_for_account.account


async def add_actor_to_account(
    account: Account, actor: Actor, name: str = "added"
) -> None:
    """Adds the actor to the account"""
    await ActorForAccount.create(account=account, actor=actor.actor_id, name=name)


async def add_actor_to_group(actor: ActorForAccount, group_name: str) -> None:
    """Adds the actor to the group"""
    await ActorGroup.create(actor=actor, name=group_name)


async def group_names_for_actor(actor: ActorForAccount) -> list[str]:
    """Returns the group names for an actor"""
    await actor.fetch_related("groups")
    return [g.name for g in actor.groups]



"""ActivityPub related functionality"""

import logging
from fastapi import APIRouter, HTTPException

from cattle_grid.activity_pub.models import StoredActivity
from cattle_grid.activity_pub.actor import is_valid_requester

from .router import ActivityPubHeaders, ActivityResponse

logger = logging.getLogger(__name__)

ap_router_object = APIRouter()


@ap_router_object.get("/object/{obj_id}", response_class=ActivityResponse)
async def return_object(obj_id, headers: ActivityPubHeaders):
    """Returns the stored activities"""

    obj = await StoredActivity.get_or_none(id=obj_id).prefetch_related("actor")

    if obj is None or not isinstance(obj.data, dict):
        raise HTTPException(404)

    if obj.data.get("id") != headers.x_ap_location:
        raise HTTPException(404)

    if not await is_valid_requester(
        headers.x_cattle_grid_requester, obj.actor, obj.data
    ):
        raise HTTPException(401)

    return obj.data



import importlib
import logging

from contextlib import asynccontextmanager
from typing import List, Callable, Awaitable, Dict
from fast_depends import inject
from fastapi import FastAPI
from faststream.rabbit import RabbitBroker

from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.extensions import Extension
from cattle_grid.model.lookup import LookupMethod, Lookup

from .util import get_transformers, transformation_steps
from .lookup import ordered_lookups
from .lifespan import iterate_lifespans, collect_lifespans

logger = logging.getLogger(__name__)


def load_extension(extension_information: dict) -> Extension:
    """Loads a single extension"""
    module_name = extension_information.get("module")

    if module_name is None:
        raise ValueError("module is required")

    module = importlib.import_module(module_name)
    extension = module.extension
    extension.configure(extension_information.get("config", {}))

    if extension.description is None and module.__doc__ is not None:
        extension.description = module.__doc__

    if "lookup_order" in extension_information:
        extension.lookup_order = extension_information["lookup_order"]
    if "api_prefix" in extension_information:
        extension.api_prefix = extension_information["api_prefix"]

    return extension


def load_extensions(settings) -> List[Extension]:
    """Loads the extensions from settings"""

    extensions = [
        load_extension(extension_information)
        for extension_information in settings.extensions
    ]

    logger.info("Loaded extensions: %s", ", ".join(f"'{e.name}'" for e in extensions))

    return extensions


def build_transformer(extensions: List[Extension]) -> Callable[[Dict], Awaitable[Dict]]:
    """Build the transformer"""
    transformers = get_transformers(extensions)
    steps = transformation_steps(transformers)

    async def transformer(data: dict):
        for step in steps:
            for plugin in step:
                data.update(await plugin.transformer(data))

        return data

    return transformer


def build_lookup(extensions: List[Extension]) -> LookupMethod:
    """Builds the lookup method"""
    methods = ordered_lookups(extensions)

    async def lookup_result(lookup: Lookup) -> Lookup:
        for method in methods:
            lookup = await inject(method)(lookup)
            if lookup.result is not None:
                return lookup
        return lookup

    return lookup_result


def set_globals(extensions: List[Extension]):
    """Sets global variables in cattle_grid.dependencies"""
    from cattle_grid.dependencies.globals import global_container

    global_container.transformer = build_transformer(extensions)
    global_container.lookup = build_lookup(extensions)


@asynccontextmanager
async def lifespan_from_extensions(extensions: List[Extension]):
    """Creates the lifespan from the extensions"""
    lifespans = collect_lifespans(extensions)

    async with iterate_lifespans(lifespans):
        yield


def add_routers_to_broker(broker: RabbitBroker, extensions: List[Extension]):
    """Adds the routers to the broker"""

    for extension in extensions:
        if extension.activity_router:
            broker.include_router(extension.activity_router)


def add_routes_to_api(app: FastAPI, extensions: List[Extension]):
    """Adds the routes to the api"""
    for extension in extensions:
        if extension.api_router:
            if extension.api_prefix:
                app.include_router(extension.api_router, prefix=extension.api_prefix)


def collect_method_information(
    extensions: List[Extension],
) -> List[MethodInformationModel]:
    """Collects the method information from the extensions"""
    return sum((extension.method_information for extension in extensions), [])



from tortoise import fields
from tortoise.models import Model


class RemoteIdentity(Model):
    """Represents the information about a remote identity"""

    id = fields.IntField(primary_key=True)
    key_id = fields.CharField(max_length=512, unique=True)
    controller = fields.CharField(max_length=512)
    public_key = fields.CharField(max_length=1024)



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from .models import Account, ActorForAccount
from .account import (
    account_with_username_password,
    create_account,
    delete_account,
    AccountAlreadyExists,
    InvalidAccountName,
    WrongPassword,
    add_permission,
    list_permissions,
    remove_permission,
    account_for_actor,
    add_actor_to_group,
    group_names_for_actor,
)


async def test_wrong_password():
    await Account.create(
        name="name",
        password_hash="$argon2id$v=19$m=65536,t=3,p=4$MIIRqgvgQbgj220jfp0MPA$YfwJSVjtjSU0zzV/P3S9nnQ/USre2wvJMjfCIjrTQbg",
    )

    result = await account_with_username_password("name", "pass")

    assert result is None


async def test_create_and_then_get():
    name = "user"
    password = "pass"

    await create_account(name, password)

    result = await account_with_username_password(name, password)

    assert result
    assert result.name == name


async def test_create_duplicate_raises_exception():
    name = "user"
    password = "pass"

    await create_account(name, password)

    with pytest.raises(AccountAlreadyExists):
        await create_account(name, password)


@pytest.mark.parametrize(
    "name", ["", "abcdefghijklmnopqrstuvwxyz", "first.second", "admin"]
)
async def test_create_name_raises_exception(name):
    with pytest.raises(InvalidAccountName):
        await create_account(name, "pass")


async def test_create_and_then_delete_wrong_password():
    name = "user"
    password = "pass"

    await create_account(name, password)

    assert 1 == await Account().filter().count()

    with pytest.raises(WrongPassword):
        await delete_account(name, "wrong")


async def test_create_and_then_delete():
    name = "user"
    password = "pass"

    await create_account(name, password)

    assert 1 == await Account().filter().count()

    await delete_account(name, password)

    assert 0 == await Account().filter().count()


async def test_add_permission():
    name = "user"
    password = "pass"

    account = await create_account(name, password)
    assert account
    await add_permission(account, "admin")
    await add_permission(account, "test")

    await account.fetch_related("permissions")

    assert set(list_permissions(account)) == {"admin", "test"}


async def test_remove_permission():
    name = "user"
    password = "pass"

    account = await create_account(name, password)
    assert account
    await add_permission(account, "admin")
    await add_permission(account, "test")

    await account.fetch_related("permissions")

    assert set(list_permissions(account)) == {"admin", "test"}

    await remove_permission(account, "admin")

    await account.fetch_related("permissions")

    assert set(list_permissions(account)) == {"test"}


async def test_account_for_actor_not_found():
    account_or_none = await account_for_actor("http://actor.example")

    assert account_or_none is None


async def test_account_for_actor():
    account = await create_account("name", "password")
    actor_id = "http://actor.example"

    await ActorForAccount.create(account=account, actor=actor_id)

    result = await account_for_actor(actor_id)

    assert result == account


async def test_account_groups():
    account = await create_account("name", "password")
    actor_id = "http://actor.example"

    actor_for_account = await ActorForAccount.create(account=account, actor=actor_id)

    await add_actor_to_group(actor_for_account, "group1")
    await add_actor_to_group(actor_for_account, "group2")

    result = await group_names_for_actor(actor_for_account)

    assert set(result) == {"group1", "group2"}



from fastapi import FastAPI
from contextlib import asynccontextmanager

import tortoise

from cattle_grid.database import database
from cattle_grid.config.auth import get_auth_config
from cattle_grid.config import load_settings
from cattle_grid.version import __version__
from .router import create_auth_router


def create_app(filenames):
    """Allows running just the auth endpoint"""

    config = load_settings(filenames)

    @asynccontextmanager
    async def lifespan(app):
        async with database(config):
            await tortoise.Tortoise.generate_schemas()
            yield

    app = FastAPI(
        lifespan=lifespan,
        title="cattle_grid.auth",
        description="""Authorization server for Fediverse applications. It basically checks HTTP Signatures for you.""",
        version=__version__,
    )

    app.include_router(create_auth_router(get_auth_config(config), tags=[]))

    return app



from cattle_grid.dependencies.globals import global_container

from .account import list_permissions
from .models import Account


async def allowed_base_urls(account: Account) -> list[str]:
    """Returns the set of base_urls the account
    is allowed to use to create an actor"""
    settings = global_container.config
    await account.fetch_related("permissions")

    permissions = list_permissions(account)

    if "admin" in permissions:
        return settings.get("frontend", {}).get("base_urls", [])

    return sum(
        (
            settings.get("permissions", {}).get(p, {}).get("base_urls", [])
            for p in permissions
        ),
        [],
    )


async def can_create_actor_at_base_url(account: Account, base_url) -> bool:
    """Checks if the account is allowed to create an actor
    at the base url"""
    allowed_urls = await allowed_base_urls(account)

    return base_url in allowed_urls



from bovine.testing import public_key, private_key

from cattle_grid.config.auth import AuthConfig

from .router import create_auth_router


async def test_create_auth_router_config():
    auth_config = AuthConfig(
        enabled=True,
        actor_id="http://localhost/actor_id",
        actor_acct_id="acct:actor@domain",
        domain_blocks={"blocked.example"},
        public_key=public_key,
        private_key=private_key,
    )
    assert create_auth_router(auth_config)



import pytest

from bovine import BovineActor

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.activity_pub.models import (
    Actor,
    PublicIdentifier,
    Follower,
    Following,
    Credential,
    ActorStatus,
)


from . import (
    create_actor,
    actor_to_object,
    compute_acct_uri,
    bovine_actor_for_actor_id,
    update_for_actor_profile,
    delete_for_actor_profile,
    delete_actor,
    remove_from_followers_following,
    DuplicateIdentifierException,
)


async def test_create_actor():
    actor = await create_actor("http://localhost/ap/")

    assert actor.actor_id.startswith("http://localhost/ap/actor/")

    assert 1 == await Actor.filter().count()


async def test_create_then_delete_actor():
    actor = await create_actor("http://localhost/ap/", preferred_username="me")
    await delete_actor(actor)

    assert 1 == await Actor.filter().count()

    await actor.refresh_from_db()
    assert actor.status == ActorStatus.deleted

    assert 0 == await PublicIdentifier.filter().count()
    assert 1 == await Credential.filter().count()


async def test_actor_to_object():
    actor = await create_actor("http://localhost/ap/")

    obj = actor_to_object(actor)

    assert obj["type"] == "Person"

    assert obj.get("followers")


async def test_update_for_actor_profile():
    actor = await create_actor("http://localhost/ap/")

    activity = update_for_actor_profile(actor)

    assert activity["type"] == "Update"
    obj = activity["object"]

    assert obj["type"] == "Person"
    assert activity["cc"]


async def test_actor_to_object_with_profile():
    actor = await create_actor(
        "http://localhost/ap/", profile={"name": "Alice Newton", "type": "Application"}
    )

    obj = actor_to_object(actor)

    assert obj["type"] == "Application"

    assert obj.get("name") == "Alice Newton"


async def test_actor_to_object_with_image():
    actor = await create_actor(
        "http://localhost/ap/", profile={"image": {"type": "Image"}}
    )

    obj = actor_to_object(actor)

    assert obj["icon"] == {"type": "Image"}


async def test_create_actor_with_identifiers():
    identifier = "acct:you@localhost"
    await create_actor("http://localhost/ap/", identifiers={"webfinger": identifier})

    pi = await PublicIdentifier.get_or_none(identifier=identifier)

    assert pi

    assert pi.name == "webfinger"
    assert pi.identifier == identifier


async def test_create_actor_with_preferred_username():
    identifier = "acct:me@localhost"
    actor = await create_actor("http://localhost/ap/", preferred_username="me")

    pi = await PublicIdentifier.get_or_none(identifier=identifier)

    assert pi

    assert pi.name == "webfinger"
    assert pi.identifier == identifier

    profile = actor_to_object(actor)
    assert profile["preferredUsername"] == "me"


def test_compute_webfinger():
    webfinger = compute_acct_uri("http://localhost/ap", "me")

    assert webfinger == "acct:me@localhost"


async def test_get_bovine_actor_not_found():
    bovine_actor = await bovine_actor_for_actor_id("http://nothing.here/actor")

    assert bovine_actor is None


async def test_get_bovine_actor():
    actor = await create_actor("http://localhost/ap/")

    bovine_actor = await bovine_actor_for_actor_id(actor.actor_id)

    assert isinstance(bovine_actor, BovineActor)


async def test_delete_for_actor_profile():
    actor = await create_actor("http://localhost/ap/")

    activity = delete_for_actor_profile(actor)

    assert activity["type"] == "Delete"

    assert activity["object"] == actor.actor_id


async def test_remove_from_followers_following():
    actor = await create_actor("http://localhost/ap/")

    remote_id = "http://remote.test"

    await Follower.create(actor=actor, follower=remote_id, accepted=True, request="")

    await Following.create(actor=actor, following=remote_id, accepted=True, request="")

    await remove_from_followers_following(remote_id)

    follower_count = await Follower.filter().count()
    following_count = await Follower.filter().count()

    assert follower_count == 0
    assert following_count == 0


async def test_identifier_ordering():
    actor = await create_actor("http://localhost/ap/")

    one = await PublicIdentifier.create(
        actor=actor, name="one", identifier="acct:jekyll@localhost", preference=1
    )
    await PublicIdentifier.create(
        actor=actor, name="two", identifier="acct:hyde@localhost", preference=100
    )

    await actor.fetch_related("identifiers")

    result = actor_to_object(actor)

    assert result["identifiers"] == [
        "acct:hyde@localhost",
        "acct:jekyll@localhost",
        actor.actor_id,
    ]
    assert result["preferredUsername"] == "hyde"

    await one.update_from_dict({"preference": 1000})
    await one.save()
    await actor.fetch_related("identifiers")

    result = actor_to_object(actor)

    assert result["identifiers"] == [
        "acct:jekyll@localhost",
        "acct:hyde@localhost",
        actor.actor_id,
    ]
    assert result["preferredUsername"] == "jekyll"


async def test_create_actor_duplicate_preferred_username():
    await create_actor("http://localhost/ap/", preferred_username="me")

    with pytest.raises(DuplicateIdentifierException):
        await create_actor("http://localhost/ap/", preferred_username="me")


async def test_actor_to_object_attachments():
    property_value_context = {
        "PropertyValue": {
            "@id": "https://schema.org/PropertyValue",
            "@context": {
                "value": "https://schema.org/value",
                "name": "https://schema.org/name",
            },
        }
    }
    property_value = {
        "type": "PropertyValue",
        "name": "key",
        "value": "value",
    }
    actor = await create_actor(
        "http://localhost/ap/",
        profile={"attachment": [property_value]},
    )

    result = actor_to_object(actor)

    assert result["attachment"] == [property_value]
    assert property_value_context in result["@context"]



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.activity_pub.models import PublicIdentifier, PublicIdentifierStatus

from .identifiers import collect_identifiers_for_actor, identifier_in_list_exists


def test_collect_identifiers_for_actor(test_actor):
    identifiers = collect_identifiers_for_actor(test_actor)

    assert identifiers == [test_actor.actor_id]


async def test_collect_identifiers_for_actor_with_acct_uri(test_actor):
    await PublicIdentifier.create(
        actor=test_actor,
        name="webfinger",
        identifier="acct:me@localhost",
        status=PublicIdentifierStatus.verified,
        preference=5,
    )

    await test_actor.fetch_related("identifiers")

    identifiers = collect_identifiers_for_actor(test_actor)

    assert identifiers == ["acct:me@localhost", test_actor.actor_id]


async def test_collect_identifiers_for_actor_with_acct_uri_unverified(test_actor):
    await PublicIdentifier.create(
        actor=test_actor,
        name="webfinger",
        identifier="acct:me@localhost",
        status=PublicIdentifierStatus.unverified,
        preference=5,
    )

    await test_actor.fetch_related("identifiers")

    identifiers = collect_identifiers_for_actor(test_actor)

    assert identifiers == [test_actor.actor_id]


@pytest.mark.parametrize(
    "identifiers, expected",
    [
        ([], False),
        (["acct:one@localhost"], True),
        (["acct:other@localhost"], False),
        (["acct:other@localhost", "acct:one@localhost"], True),
    ],
)
async def test_identifier_in_list_exists(test_actor, identifiers, expected):
    await PublicIdentifier.create(
        actor=test_actor,
        name="webfinger",
        identifier="acct:one@localhost",
        status=PublicIdentifierStatus.unverified,
        preference=5,
    )
    await PublicIdentifier.create(
        actor=test_actor,
        name="webfinger",
        identifier="acct:two@localhost",
        status=PublicIdentifierStatus.unverified,
        preference=5,
    )
    result = await identifier_in_list_exists(identifiers)

    assert result == expected



from behave import when, then
from behave.api.async_step import async_run_until_complete

from bovine.activitystreams import factories_for_actor_object

from cattle_grid.testing.features import publish_as, fetch_request


@when('"{alice}" publishes a "{moo}" animal sound to her follows')
@async_run_until_complete
async def send_sound(context, alice, moo):
    for connection in context.connections.values():
        await connection.clear_incoming()

    alice_actor = context.actors[alice]
    activity_factory, _ = factories_for_actor_object(alice_actor)

    activity = (
        activity_factory.custom(type="AnimalSound", content="moo").as_public().build()
    )

    await publish_as(
        context,
        alice,
        "publish_activity",
        {"actor": context.actors[alice].get("id"), "data": activity},
    )


@then('"{bob}" can retrieve the activity')
@async_run_until_complete
async def can_retrieve_activity(context, bob):
    result = await fetch_request(context, bob, context.activity.get("id"))
    assert result



from dataclasses import dataclass
from typing import Literal
import logging


from bovine import BovineActor
from bovine.crypto.types import CryptographicIdentifier
from fediverse_pasture.server.utils import actor_object_to_public_key

from .model import RemoteIdentity

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@dataclass
class PublicKeyCache:
    """Caches public keys in the database and fetches them
    using bovine_actor"""

    bovine_actor: BovineActor
    """used to fetch the public key"""

    async def cryptographic_identifier(
        self, key_id: str
    ) -> CryptographicIdentifier | Literal["gone"] | None:
        """Returns "gone" if Tombstone

        :param key_id: URI of the public key to fetch
        :returns:
        """

        try:
            result = await self.bovine_actor.get(key_id)

            if result is None:
                return "gone"

            if result.get("type") == "Tombstone":
                logger.info("Got Tombstone for %s", key_id)
                return "gone"

            public_key, owner = actor_object_to_public_key(result, key_id)

            if public_key is None or owner is None:
                return None

            return CryptographicIdentifier.from_pem(public_key, owner)
        except Exception as e:
            logger.info("Failed to fetch public key for %s with %s", key_id, repr(e))
            # logger.exception(e)
            return None

    async def from_cache(self, key_id: str) -> CryptographicIdentifier | None:
        identity = await RemoteIdentity.get_or_none(key_id=key_id)

        if identity is None:
            identifier = await self.cryptographic_identifier(key_id)
            if identifier is None:
                return None

            if identifier == "gone":
                await RemoteIdentity.create(
                    key_id=key_id, public_key="gone", controller="gone"
                )
                return None

            try:
                controller, public_key = identifier.as_tuple()
                identity = await RemoteIdentity.create(
                    key_id=key_id, public_key=public_key, controller=controller
                )
                await identity.save()
            except Exception as e:
                logger.exception(e)
            return identifier

        if identity.controller == "gone":
            return None

        return CryptographicIdentifier.from_tuple(
            identity.controller, identity.public_key
        )



from unittest.mock import AsyncMock

from cattle_grid.testing import mocked_config
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model.account import TriggerMessage

from cattle_grid.account.account import add_actor_to_group
from cattle_grid.account.models import ActorForAccount


from .trigger import handle_trigger


async def test_handle_trigger(test_actor_with_account):
    broker = AsyncMock()
    actor_for_account = await ActorForAccount.get(
        actor=test_actor_with_account.actor_id
    )
    assert actor_for_account

    await handle_trigger(
        TriggerMessage(
            actor=test_actor_with_account.actor_id,  # type:ignore
        ),
        actor=actor_for_account,
        broker=broker,
        correlation_id="uuid",
        method="method",
    )

    broker.publish.assert_awaited_once()

    (_, kwargs) = broker.publish.call_args

    assert kwargs["routing_key"] == "method"
    assert kwargs["correlation_id"] == "uuid"


async def test_handle_trigger_with_rewrite(test_actor_with_account):
    broker = AsyncMock()
    actor_for_account = await ActorForAccount.get(
        actor=test_actor_with_account.actor_id
    )
    assert actor_for_account

    await add_actor_to_group(actor_for_account, "group")

    config = {
        "rewrite": {
            "group": {
                "method": "changed",
            }
        }
    }

    with mocked_config(config):
        await handle_trigger(
            TriggerMessage(
                actor=test_actor_with_account.actor_id,
            ),
            actor=actor_for_account,
            broker=broker,
            correlation_id="uuid",
            method="method",
        )

    broker.publish.assert_awaited_once()

    (_, kwargs) = broker.publish.call_args

    assert kwargs["routing_key"] == "changed"
    assert kwargs["correlation_id"] == "uuid"



from .auth.model import RemoteIdentity
from .activity_pub.models import Actor
from .database import database


async def statistics(config):
    async with database(config=config):
        remote_identity_count = await RemoteIdentity.filter().count()
        actor_count = await Actor.filter().count()

        print(f"Remote identity count: {remote_identity_count:10d}")
        print(f"Actor count:           {actor_count:10d}")



import pytest

from cattle_grid.testing.fixtures import database_for_tests  # noqa
from .models import Follower, Blocking


from .actor import create_actor, is_valid_requester


@pytest.fixture
async def test_actor():
    actor = await create_actor("http://localhost/ap/")

    await Follower.create(
        actor=actor, follower="http://follower.test", accepted=True, request="xxx"
    )
    await Blocking.create(
        actor=actor, blocking="http://blocking.test", active=True, request="xxx"
    )

    return actor


@pytest.mark.parametrize(
    "obj,expected",
    [
        ({}, False),
        ({"to": ["http://remote.example"]}, True),
        ({"to": ["as:Public"]}, True),
    ],
)
async def test_is_valid_requester_remote(test_actor, obj, expected):
    valid = await is_valid_requester("http://remote.example", test_actor, obj)

    assert valid == expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        ({}, False),
        ({"to": ["http://blocking.test"]}, False),
        ({"to": ["as:Public"]}, False),
    ],
)
async def test_is_valid_requester_blocking(test_actor, obj, expected):
    valid = await is_valid_requester("http://blocking.test", test_actor, obj)

    assert valid == expected



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "account" ADD "meta_information" JSONB NOT NULL DEFAULT '{}';"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "account" DROP COLUMN "meta_information";"""



from tortoise import BaseDBAsyncClient


async def upgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "actorforaccount" ADD "status" VARCHAR(7)NOT NULL DEFAULT 'active';"""


async def downgrade(db: BaseDBAsyncClient) -> str:
    return """
        ALTER TABLE "actorforaccount" DROP COLUMN "status";"""



from typing import List
from cattle_grid.extensions import Extension
from cattle_grid.model.lookup import LookupMethod


def ordered_lookups(extensions: List[Extension]) -> List[LookupMethod]:
    """Returns a list of LookupMethod ordered by lookup order"""
    sorted_extensions = sorted(
        extensions, key=lambda extension: extension.lookup_order or 0, reverse=True
    )

    return [
        extension.lookup_method
        for extension in sorted_extensions
        if extension.lookup_method
    ]



from unittest.mock import AsyncMock

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.testing import mocked_config

from cattle_grid.activity_pub.actor import actor_to_object
from cattle_grid.activity_pub.models import PublicIdentifier

from cattle_grid.model.exchange import (
    UpdateActorMessage,
    UpdateIdentifierAction,
    UpdateActionType,
)
from .handlers import update_actor


async def test_update_actor(test_actor):
    new_name = "Alyssa Newton"
    msg = UpdateActorMessage(actor=test_actor.actor_id, profile={"name": new_name})

    await update_actor(msg, test_actor, AsyncMock())

    await test_actor.refresh_from_db()

    obj = actor_to_object(test_actor)

    assert obj.get("name") == new_name


async def test_update_actor_profile_not_overwritten(test_actor):
    new_name = "Alyssa Newton"
    new_description = "I was originally called Leibnitz. Also I'm a cat."
    msg = UpdateActorMessage(actor=test_actor.actor_id, profile={"name": new_name})

    await update_actor(msg, test_actor, AsyncMock())
    await test_actor.refresh_from_db()

    obj = actor_to_object(test_actor)
    assert obj.get("name") == new_name

    msg = UpdateActorMessage(
        actor=test_actor.actor_id, profile={"summary": new_description}
    )

    await update_actor(msg, test_actor, AsyncMock())
    await test_actor.refresh_from_db()

    obj = actor_to_object(test_actor)
    assert obj.get("name") == new_name


async def test_create_identifier(test_actor_with_account):
    with mocked_config({"frontend": {"base_urls": ["http://localhost"]}}):
        identifier = "acct:new@localhost"
        new_identifier = UpdateIdentifierAction(
            action=UpdateActionType.create_identifier,
            identifier=identifier,
            primary=False,
        )

        msg = UpdateActorMessage(
            actor=test_actor_with_account.actor_id, actions=[new_identifier]
        )

        await update_actor(msg, test_actor_with_account, AsyncMock())

        await test_actor_with_account.refresh_from_db()
        await test_actor_with_account.fetch_related("identifiers")

        obj = actor_to_object(test_actor_with_account)

        assert identifier in obj.get("identifiers", [])


async def test_update_identifier(test_actor):
    identifier = "acct:one@localhost"
    await PublicIdentifier.create(
        actor=test_actor,
        identifier="acct:two@localhost",
        name="through_exchange",
        preference=1,
    )
    await PublicIdentifier.create(
        actor=test_actor,
        identifier=identifier,
        name="through_exchange",
        preference=0,
    )

    await test_actor.refresh_from_db()
    await test_actor.fetch_related("identifiers")

    obj = actor_to_object(test_actor)

    assert obj["preferredUsername"] == "two"

    update_identifier = UpdateIdentifierAction(
        action=UpdateActionType.update_identifier,
        identifier=identifier,
        primary=True,
    )

    msg = UpdateActorMessage(actor=test_actor.actor_id, actions=[update_identifier])

    await update_actor(msg, test_actor, AsyncMock())

    await test_actor.refresh_from_db()
    await test_actor.fetch_related("identifiers")

    obj = actor_to_object(test_actor)

    assert obj["preferredUsername"] == "one"
